@@ -3,7 +3,12 @@
 
 #include "btllib/nthash.hpp"
 
+// clang-format off
+// NOLINTBEGIN llvm-include-order
+#include <limits>
 #include "cpptoml.h"
+// NOLINTEND llvm-include-order
+// clang-format on
 
 #include <atomic>
 #include <climits>
@@ -600,7 +605,8 @@ public:
   const std::vector<std::string>& get_seeds() const { return seeds; }
   /** Get the seeds used in parsed format. Parsed format is a vector of indices
    * of 0s in the seed. */
-  const std::vector<SpacedSeed>& get_parsed_seeds() const
+  const std::vector<btllib::hashing_internals::SpacedSeed>& get_parsed_seeds()
+    const
   {
     return parsed_seeds;
   }
@@ -641,7 +647,7 @@ private:
   SeedBloomFilter(const std::shared_ptr<BloomFilterInitializer>& bfi);
 
   std::vector<std::string> seeds;
-  std::vector<SpacedSeed> parsed_seeds;
+  std::vector<btllib::hashing_internals::SpacedSeed> parsed_seeds;
   KmerBloomFilter kmer_bloom_filter;
 };
 
