@@ -50,11 +50,14 @@ public:
     static const unsigned SHORT_MODE = 32;
     /** Optimizes performance for long sequences (approx. >5kbp) */
     static const unsigned LONG_MODE = 64;
+    /** Include read sequence Phred score along with minimizer information. */
+    static const unsigned QUAL = 128;
   };
 
   bool output_id() const { return bool(~flags & Flag::NO_ID); }
   bool output_bx() const { return bool(flags & Flag::BX); }
   bool output_seq() const { return bool(flags & Flag::SEQ); }
+  bool output_qual() const { return bool(flags & Flag::QUAL); }
   bool filter_in() const { return bool(flags & Flag::FILTER_IN); }
   bool filter_out() const { return bool(flags & Flag::FILTER_OUT); }
   bool short_mode() const { return bool(flags & Flag::SHORT_MODE); }
@@ -77,10 +80,26 @@ public:
     {
     }
 
+    Minimizer(uint64_t min_hash,
+              uint64_t out_hash,
+              size_t pos,
+              bool forward,
+              std::string seq,
+              std::string qual)
+      : min_hash(min_hash)
+      , out_hash(out_hash)
+      , pos(pos)
+      , forward(forward)
+      , seq(std::move(seq))
+      , qual(std::move(qual))
+    {
+    }
+
     uint64_t min_hash = 0, out_hash = 0;
     size_t pos = 0;
     bool forward = false;
     std::string seq;
+    std::string qual;
   };
 
   using HashedKmer = Minimizer;
@@ -127,6 +146,7 @@ public:
    * stdin.
    * @param k k-mer size for the minimizer.
    * @param w window size when selecting minimizers.
+   * @param q quality threshold to ignore potential minimizers.
    * @param flags Modifier flags. Specifiying either short or long mode flag is
    * mandatory; other flags are optional.
    * @param threads Maximum number of processing threads to use. Must be at
@@ -147,6 +167,16 @@ public:
           const btllib::BloomFilter& bf1 = Indexlr::dummy_bf(),
           const btllib::BloomFilter& bf2 = Indexlr::dummy_bf());
 
+  Indexlr(std::string seqfile,
+          size_t k,
+          size_t w,
+          size_t q,
+          unsigned flags = 0,
+          unsigned threads = 5,
+          bool verbose = false,
+          const btllib::BloomFilter& bf1 = Indexlr::dummy_bf(),
+          const btllib::BloomFilter& bf2 = Indexlr::dummy_bf());
+
   ~Indexlr();
 
   void close() noexcept;
@@ -199,6 +229,12 @@ private:
                                  bool filter_out,
                                  const BloomFilter& filter_in_bf,
                                  const BloomFilter& filter_out_bf);
+
+  static void filter_kmer_qual(Indexlr::HashedKmer& hk,
+                               const std::string& kmer_qual,
+                               size_t q);
+  static size_t calc_kmer_quality(const std::string& qual);
+
   static void calc_minimizer(
     const std::vector<Indexlr::HashedKmer>& hashed_kmers_buffer,
     const Indexlr::Minimizer*& min_current,
@@ -208,10 +244,12 @@ private:
     ssize_t& min_pos_prev,
     size_t w,
     std::vector<Indexlr::Minimizer>& minimizers);
-  std::vector<Minimizer> minimize(const std::string& seq) const;
+  std::vector<Minimizer> minimize(const std::string& seq,
+                                  const std::string& qual) const;
 
   const std::string seqfile;
   const size_t k, w;
+  size_t q;
   const unsigned flags;
   const bool verbose;
   const long id;
@@ -296,9 +334,11 @@ private:
   bool last_block_num_valid = false;
 };
 
+// Constructor for Indexlr class when q is specified
 inline Indexlr::Indexlr(std::string seqfile,
                         const size_t k,
                         const size_t w,
+                        const size_t q,
                         const unsigned flags,
                         const unsigned threads,
                         const bool verbose,
@@ -307,6 +347,7 @@ inline Indexlr::Indexlr(std::string seqfile,
   : seqfile(std::move(seqfile))
   , k(k)
   , w(w)
+  , q(q)
   , flags(flags)
   , verbose(verbose)
   , id(++last_id())
@@ -335,6 +376,19 @@ inline Indexlr::Indexlr(std::string seqfile,
   }
 }
 
+// Constructor for Indexlr class when q is not specified
+inline Indexlr::Indexlr(std::string seqfile,
+                        const size_t k,
+                        const size_t w,
+                        const unsigned flags,
+                        const unsigned threads,
+                        const bool verbose,
+                        const BloomFilter& bf1,
+                        const BloomFilter& bf2)
+  : Indexlr(std::move(seqfile), k, w, 0, flags, threads, verbose, bf1, bf2)
+{
+}
+
 inline Indexlr::~Indexlr()
 {
   close();
@@ -431,6 +485,34 @@ Indexlr::filter_hashed_kmer(Indexlr::HashedKmer& hk,
   }
 }
 
+inline void
+Indexlr::filter_kmer_qual(Indexlr::HashedKmer& hk,
+                          const std::string& kmer_qual,
+                          size_t q)
+{
+  if (calc_kmer_quality(kmer_qual) < q) {
+    hk.min_hash = std::numeric_limits<uint64_t>::max();
+  }
+}
+
+inline size_t
+Indexlr::calc_kmer_quality(const std::string& qual)
+{
+  // convert the quality scores to integers
+  std::vector<int> qual_ints;
+  const int thirty_three = 33;
+  qual_ints.reserve(qual.size());
+  for (auto c : qual) {
+    qual_ints.push_back(c - thirty_three);
+  }
+  // calculate the mean (potential improvement: use other statistics)
+  size_t sum = 0;
+  for (auto q : qual_ints) {
+    sum += q;
+  }
+  return (sum / qual_ints.size());
+}
+
 inline void
 Indexlr::calc_minimizer(
   const std::vector<Indexlr::HashedKmer>& hashed_kmers_buffer,
@@ -469,7 +551,7 @@ Indexlr::calc_minimizer(
 }
 
 inline std::vector<Indexlr::Minimizer>
-Indexlr::minimize(const std::string& seq) const
+Indexlr::minimize(const std::string& seq, const std::string& qual) const
 {
   if ((k > seq.size()) || (w > seq.size() - k + 1)) {
     return {};
@@ -486,12 +568,17 @@ Indexlr::minimize(const std::string& seq) const
     hk = HashedKmer(nh.hashes()[0],
                     nh.hashes()[1],
                     nh.get_pos(),
-                    nh.forward(),
-                    output_seq() ? seq.substr(nh.get_pos(), k) : "");
+                    nh.get_forward_hash() <= nh.get_reverse_hash(),
+                    output_seq() ? seq.substr(nh.get_pos(), k) : "",
+                    output_qual() ? qual.substr(nh.get_pos(), k) : "");
 
     filter_hashed_kmer(
       hk, filter_in(), filter_out(), filter_in_bf.get(), filter_out_bf.get());
 
+    if (q > 0) {
+      filter_kmer_qual(hk, qual.substr(nh.get_pos(), k), q);
+    }
+
     if (idx + 1 >= w) {
       calc_minimizer(hashed_kmers_buffer,
                      min_current,
@@ -511,7 +598,8 @@ Indexlr::read()
 {
   if (ready_blocks_owners()[id % MAX_SIMULTANEOUS_INDEXLRS] != id) {
     ready_blocks_array()[id % MAX_SIMULTANEOUS_INDEXLRS] =
-      std::unique_ptr<decltype(output_queue)::Block>(
+      std::unique_ptr< // NOLINT(modernize-make-unique)
+        decltype(output_queue)::Block>(
         new decltype(output_queue)::Block(reader.get_block_size()));
     ready_blocks_owners()[id % MAX_SIMULTANEOUS_INDEXLRS] = id;
     ready_blocks_current()[id % MAX_SIMULTANEOUS_INDEXLRS] = 0;
@@ -549,13 +637,14 @@ Indexlr::Worker::work()
       Record record;
       auto& reader_record = input_block.data[idx];
       record.num = reader_record.num;
+      if (indexlr.output_bx()) {
+        record.barcode =
+          indexlr.extract_barcode(reader_record.id, reader_record.comment);
+      }
       if (indexlr.output_id()) {
         record.id = std::move(reader_record.id);
       }
-      if (indexlr.output_bx()) {
-        record.barcode =
-          indexlr.extract_barcode(record.id, reader_record.comment);
-      }
+
       record.readlen = reader_record.seq.size();
 
       check_info(indexlr.verbose && indexlr.k > record.readlen,
@@ -582,7 +671,8 @@ Indexlr::Worker::work()
 
       if (indexlr.k <= record.readlen &&
           indexlr.w <= record.readlen - indexlr.k + 1) {
-        record.minimizers = indexlr.minimize(reader_record.seq);
+        record.minimizers =
+          indexlr.minimize(reader_record.seq, reader_record.qual);
       } else {
         record.minimizers = {};
       }
@@ -615,4 +705,4 @@ Indexlr::Worker::work()
 
 } // namespace btllib
 
-#endif
\ No newline at end of file
+#endif
